#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include <commctrl.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "StackFunc.h"
#include "StrFunc.h"
#include "Resources\Resource.h"


//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#include "StackFunc.h"

//Include stack functions
#define WideCharUpper
#define xstrcmpinW
#define xstrcmpiW
#define xstrcpyW
#define xstrcpynW
#include "StrFunc.h"

//Defines
#define STRID_HOTKEYS          1
#define STRID_COMPLETEWITHLIST 2
#define STRID_COMPLETENEXT     3
#define STRID_COMPLETEPREV     4
#define STRID_AUTOLIST         5
#define STRID_ENABLE           6
#define STRID_AFTERCHAR_PRE    7
#define STRID_AFTERCHAR_POST   8
#define STRID_RIGHTDELIMITERS  9
#define STRID_OK               10
#define STRID_CANCEL           11

#define DLLA_AUTOCOMPLETE_SET  1

#define BUFFER_SIZE      1024

#define OF_SETTINGS       0x1
#define OF_WINDOWRECT     0x2

#define FIRST_NONLATIN   128

#define SIZE_MINX      30
#define SIZE_MINY      30

//CreateAutoCompleteWindow flags
#define CAW_COMPLETEONE    0x1
#define CAW_COMPLETEEXACT  0x2
#define CAW_COMPLETEWINDOW 0x4
#define CAW_AUTOLIST       0x8

#define AUTOCOMPLETEA   "AutoComplete"
#define AUTOCOMPLETEW  L"AutoComplete"
#define IDC_LIST        1001

#ifndef WM_MOUSEWHEEL
  #define WM_MOUSEWHEEL 0x020A
#endif

typedef struct _EXTENSIONINFO {
  struct _EXTENSIONINFO *next;
  struct _EXTENSIONINFO *prev;
  wchar_t *wpExtension;
  int nExtensionLen;
} EXTENSIONINFO;

typedef struct _DELIMITERINFO {
  struct _DELIMITERINFO *next;
  struct _DELIMITERINFO *prev;
  wchar_t wchDelimiter;
} DELIMITERINFO;

typedef struct _VARIABLEINFO {
  struct _VARIABLEINFO *next;
  struct _VARIABLEINFO *prev;
  wchar_t *wpVarName;
  int nVarNameLen;
  wchar_t *wpVarValue;
  int nVarValueLen;
} VARIABLEINFO;

typedef struct _WORDINFO {
  struct _WORDINFO *next;
  struct _WORDINFO *prev;
  wchar_t *wpWord;
  int nWordLen;
} WORDINFO;

typedef struct _BLOCKINFO {
  struct _BLOCKINFO *next;
  struct _BLOCKINFO *prev;
  wchar_t wchFirstUpperChar;
  wchar_t *wpWord;
  int nWordLen;

  struct _BLOCKINFO *master;
  wchar_t *wpBlock;
  int nBlockLen;
  int nLinesInBlock;
  HSTACK hHotSpotStack;
  int nHotSpotBlockBegin;
} BLOCKINFO;

typedef struct _HOTSPOT {
  struct _HOTSPOT *next;
  struct _HOTSPOT *prev;
  int nHotSpotInitPos;
  int nHotSpotInitLen;
  int nHotSpotPos;
  int nHotSpotLen;
} HOTSPOT;

typedef struct _HEXTENSIONSTACK {
  int first;
  int last;
  int lpExtensionLens[MAX_PATH];
} HEXTENSIONSTACK;

typedef struct _HBLOCKSTACK {
  int first;
  int last;
  int lpSorted[FIRST_NONLATIN + 1];
} HBLOCKSTACK;

typedef struct _HTHEMESTACK {
  struct _HTHEMESTACK *next;
  struct _HTHEMESTACK *prev;
  HEXTENSIONSTACK hExtensionStack;
  HSTACK hDelimiterStack;
  HSTACK hVarStack;
  HBLOCKSTACK hBlockStack;
} HTHEMESTACK;

typedef struct _MANUALSET {
  struct _MANUALSET *next;
  struct _MANUALSET *prev;
  HWND hWnd;
  wchar_t wszExt[MAX_PATH];
} MANUALSET;

//// Functions prototypes
BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK GetMsgProc(int code, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK CompleteWithListProc(void *lpParameter);
BOOL CALLBACK CompleteNextProc(void *lpParameter);
BOOL CALLBACK CompletePrevProc(void *lpParameter);
LRESULT CALLBACK AutoCompleteWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewListboxProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

//AutoComplete window
BOOL CreateAutoCompleteWindow(HTHEMESTACK *lpTheme, DWORD dwFlags);
BOOL MoveAutoCompleteWindow();
void CloseAutoCompleteWindow();

//AutoComplete window listbox
void FillListbox(HBLOCKSTACK *hBlockStack);
int FindStringListbox(wchar_t *wpWord);
void SetSelListbox(int nIndex);
BLOCKINFO* GetBlockListbox();

//Scheme
HTHEMESTACK* StackLoadSyntaxFile(HANDLE hFile, HSTACK *hSyntaxFilesStack);
int ParseBlock(HTHEMESTACK *lpScheme, HSTACK *hHotSpotStack, const wchar_t *wpInput, int nInputLen, wchar_t *wszOutput, int *nOutputLines);
HTHEMESTACK* StackInsertScheme(HSTACK *hStack);
HTHEMESTACK* StackGetScheme(HSTACK *hStack, const wchar_t *wpExtension, int nExtensionLen);
HTHEMESTACK* StackGetCurrentScheme(HSTACK *hStack);
void StackFreeScheme(HSTACK *hStack);

//Extension
EXTENSIONINFO* StackInsertExtension(HEXTENSIONSTACK *hStack, int nExtensionLen);
EXTENSIONINFO* StackGetExtension(HEXTENSIONSTACK *hStack, const wchar_t *wpExtension, int nExtensionLen);
void StackFreeExtension(HEXTENSIONSTACK *hStack);

//Delimiter
DELIMITERINFO* StackInsertDelimiter(HSTACK *hStack);
DELIMITERINFO* StackGetDelimiter(HSTACK *hStack, wchar_t wchDelimiter);
void StackFreeDelimiter(HSTACK *hStack);

//Variable
VARIABLEINFO* StackInsertVariable(HSTACK *hStack);
VARIABLEINFO* StackGetVariable(HSTACK *hStack, const wchar_t *wpVarName, int nVarNameLen);
void StackFreeVariable(HSTACK *hStack);

//Word
WORDINFO* StackInsertWord(HSTACK *hStack);
void StackFreeWord(HSTACK *hStack);

//Word part
BOOL GetEditWordPart(HSTACK *hDelimiterStack, wchar_t *wszWord, int nWordMax, int *nMin, int *nMax);
void CompleteWordPart(BLOCKINFO *lpBlockInfo, int nMin, int nMax);

//Block
BLOCKINFO* StackInsertBlock(HBLOCKSTACK *hBlockStack, wchar_t *wpWord, int nWordLen);
BLOCKINFO* StackGetBlock(HBLOCKSTACK *hBlockStack, const wchar_t *wpWordPart, int nWordPartLen, BOOL *bOnlyOne);
void StackFreeBlock(HBLOCKSTACK *hBlockStack);

//Hot spot
HOTSPOT* StackInsertHotSpot(HSTACK *hStack, int nHotSpotPos);
HOTSPOT* GetHotSpot(BLOCKINFO *lpBlockInfo, HWND hWnd, CHARRANGE *cr);
BOOL NextHotSpot(BLOCKINFO *lpBlockInfo, HWND hWnd, BOOL bPrev);
void StackResetHotSpot(BLOCKINFO *lpBlockInfo);
void StackFreeHotSpot(HSTACK *hStack);

//Manual extension set
MANUALSET* StackInsertManual(HSTACK *hStack);
MANUALSET* StackGetManual(HSTACK *hStack, HWND hWnd);
void StackDeleteManual(HSTACK *hStack, void *ms);

//Other
int GetCharHeight(HWND hWnd);
void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt);
int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote);
int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote);
BOOL NextLineA(char **pText);
BOOL NextLineW(wchar_t **wpText);
BOOL SkipCommentA(char **pText);
BOOL SkipCommentW(wchar_t **wpText);
char* GetFileExtA(char *pFile);
wchar_t* GetFileExtW(wchar_t *wpFile);
void ReadSyntaxFilesA();
void ReadSyntaxFilesW();
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//// Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hWndEdit;
HICON hMainIcon;
BOOL bOldWindows;
BOOL bOldRichEdit;
BOOL bAkelEdit;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
char szAutoCompleteDir[MAX_PATH];
wchar_t wszAutoCompleteDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
HTHEMESTACK *lpWindowTheme=NULL;
BLOCKINFO *lpCurrentBlockElement=NULL;
int nWindowBlockBegin=0;
int nWindowBlockEnd=0;
RECT rcAutoComplete={0, 0, 140, 102};
HWND hWndAutoComplete=NULL;
HWND hWndListBox=NULL;
HFONT hListboxFont=NULL;
WNDPROC OldListBoxProc=NULL;
HHOOK hHook=NULL;
BOOL bRightDelimitersEnable=TRUE;
BOOL bAutoListEnable=FALSE;
int nAutoListAfter=2;
BOOL bAutoCompleteWindowRectSave=FALSE;
HSTACK hSyntaxFilesStack={0};
HSTACK hManualStack={0};
BOOL bReadSyntaxFiles=FALSE;
DWORD dwCompleteWithList=544; //"Ctrl+Space"
DWORD dwCompleteNext=0;       //"None"
DWORD dwCompletePrev=0;       //"None"
PLUGINFUNCTIONA *pfaCompleteWithList=NULL;
PLUGINFUNCTIONA *pfaCompleteNext=NULL;
PLUGINFUNCTIONA *pfaCompletePrev=NULL;
PLUGINFUNCTIONW *pfwCompleteWithList=NULL;
PLUGINFUNCTIONW *pfwCompleteNext=NULL;
PLUGINFUNCTIONW *pfwCompletePrev=NULL;
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;
WNDPROCDATA *NewEditProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="AutoComplete";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (pd->lParam)
  {
    if (bInitMain)
    {
      if (IsExtCallParamValid(pd->lParam, 1))
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOCOMPLETE_SET)
        {
          MANUALSET *ms;
          wchar_t *wpExt=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpExt=(wchar_t *)GetExtCallParam(pd->lParam, 2);

          if (wpExt)
          {
            if (pd->hWndEdit)
            {
              if (!(ms=StackGetManual(&hManualStack, pd->hWndEdit)))
                ms=StackInsertManual(&hManualStack);

              if (ms)
              {
                ms->hWnd=pd->hWndEdit;
                xstrcpynW(ms->wszExt, wpExt, MAX_PATH);
              }
            }
          }
        }
      }
    }
    else return;
  }
  else
  {
    if (!bInitCommon) InitCommon(pd);
    if (!bInitMain) InitMain();

    if (!pd->bOnStart)
    {
      if (bOldWindows)
        DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_MAIN), hMainWnd, (DLGPROC)MainDlgProcA);
      else
        DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_MAIN), hMainWnd, (DLGPROC)MainDlgProcW);
    }
  }

  //Stay in memory, and show as active
  pd->nUnload=UD_NONUNLOAD_ACTIVE;
}

BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndCompleteWithList;
  static HWND hWndCompleteNext;
  static HWND hWndCompletePrev;
  static HWND hWndAutoListEnable;
  static HWND hWndAutoListAfter;
  static HWND hWndAutoListAfterSpin;
  static HWND hWndRightDelimitersEnable;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndCompleteWithList=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETEWITHLIST);
    hWndCompleteNext=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETENEXT);
    hWndCompletePrev=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETEPREV);
    hWndAutoListEnable=GetDlgItem(hDlg, IDC_MAIN_AUTOLIST_ENABLE);
    hWndAutoListAfter=GetDlgItem(hDlg, IDC_MAIN_AUTOLISTAFTER);
    hWndAutoListAfterSpin=GetDlgItem(hDlg, IDC_MAIN_AUTOLISTAFTER_SPIN);
    hWndRightDelimitersEnable=GetDlgItem(hDlg, IDC_MAIN_RIGHTDELIMITERS_ENABLE);

    SetDlgItemTextA(hDlg, IDC_MAIN_HOTKEY_GROUP, GetLangStringA(wLangSystem, STRID_HOTKEYS));
    SetDlgItemTextA(hDlg, IDC_MAIN_HOTKEY_COMPLETEWITHLIST_LABEL, GetLangStringA(wLangSystem, STRID_COMPLETEWITHLIST));
    SetDlgItemTextA(hDlg, IDC_MAIN_HOTKEY_COMPLETENEXT_LABEL, GetLangStringA(wLangSystem, STRID_COMPLETENEXT));
    SetDlgItemTextA(hDlg, IDC_MAIN_HOTKEY_COMPLETEPREV_LABEL, GetLangStringA(wLangSystem, STRID_COMPLETEPREV));
    SetDlgItemTextA(hDlg, IDC_MAIN_AUTOLIST_GROUP, GetLangStringA(wLangSystem, STRID_AUTOLIST));
    SetDlgItemTextA(hDlg, IDC_MAIN_AUTOLIST_ENABLE, GetLangStringA(wLangSystem, STRID_ENABLE));
    SetDlgItemTextA(hDlg, IDC_MAIN_AUTOLISTAFTER_PRELABEL, GetLangStringA(wLangSystem, STRID_AFTERCHAR_PRE));
    SetDlgItemTextA(hDlg, IDC_MAIN_AUTOLISTAFTER_POSTLABEL, GetLangStringA(wLangSystem, STRID_AFTERCHAR_POST));
    SetDlgItemTextA(hDlg, IDC_MAIN_RIGHTDELIMITERS_ENABLE, GetLangStringA(wLangSystem, STRID_RIGHTDELIMITERS));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    if (bAutoListEnable) SendMessage(hWndAutoListEnable, BM_SETCHECK, BST_CHECKED, 0);
    SendMessage(hWndAutoListAfterSpin, UDM_SETBUDDY, (WPARAM)hWndAutoListAfter, 0);
    SendMessage(hWndAutoListAfterSpin, UDM_SETRANGE, 0, MAKELONG(99, 1));
    SetDlgItemInt(hDlg, IDC_MAIN_AUTOLISTAFTER, nAutoListAfter, FALSE);

    if (bRightDelimitersEnable) SendMessage(hWndRightDelimitersEnable, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndCompleteWithList, HKM_SETHOTKEY, dwCompleteWithList, 0);
    SendMessage(hWndCompleteNext, HKM_SETHOTKEY, dwCompleteNext, 0);
    SendMessage(hWndCompletePrev, HKM_SETHOTKEY, dwCompletePrev, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompleteWithList, pfaCompleteWithList?dwCompleteWithList:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompleteNext, pfaCompleteNext?dwCompleteNext:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompletePrev, pfaCompletePrev?dwCompletePrev:0);

    PostMessage(hDlg, WM_COMMAND, IDC_MAIN_AUTOLIST_ENABLE, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_MAIN_AUTOLIST_ENABLE)
    {
      bAutoListEnable=SendMessage(hWndAutoListEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndAutoListAfter, bAutoListEnable);
      EnableWindow(hWndAutoListAfterSpin, bAutoListEnable);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      dwCompleteWithList=(WORD)SendMessage(hWndCompleteWithList, HKM_GETHOTKEY, 0, 0);
      dwCompleteNext=(WORD)SendMessage(hWndCompleteNext, HKM_GETHOTKEY, 0, 0);
      dwCompletePrev=(WORD)SendMessage(hWndCompletePrev, HKM_GETHOTKEY, 0, 0);

      nAutoListAfter=GetDlgItemInt(hDlg, IDC_MAIN_AUTOLISTAFTER, NULL, FALSE);

      bRightDelimitersEnable=SendMessage(hWndRightDelimitersEnable, BM_GETCHECK, 0, 0);

      SaveOptionsA(OF_SETTINGS);
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        UninitMain();
        InitMain();
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndCompleteWithList;
  static HWND hWndCompleteNext;
  static HWND hWndCompletePrev;
  static HWND hWndAutoListEnable;
  static HWND hWndAutoListAfter;
  static HWND hWndAutoListAfterSpin;
  static HWND hWndRightDelimitersEnable;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndCompleteWithList=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETEWITHLIST);
    hWndCompleteNext=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETENEXT);
    hWndCompletePrev=GetDlgItem(hDlg, IDC_MAIN_HOTKEY_COMPLETEPREV);
    hWndAutoListEnable=GetDlgItem(hDlg, IDC_MAIN_AUTOLIST_ENABLE);
    hWndAutoListAfter=GetDlgItem(hDlg, IDC_MAIN_AUTOLISTAFTER);
    hWndAutoListAfterSpin=GetDlgItem(hDlg, IDC_MAIN_AUTOLISTAFTER_SPIN);
    hWndRightDelimitersEnable=GetDlgItem(hDlg, IDC_MAIN_RIGHTDELIMITERS_ENABLE);

    SetDlgItemTextW(hDlg, IDC_MAIN_HOTKEY_GROUP, GetLangStringW(wLangSystem, STRID_HOTKEYS));
    SetDlgItemTextW(hDlg, IDC_MAIN_HOTKEY_COMPLETEWITHLIST_LABEL, GetLangStringW(wLangSystem, STRID_COMPLETEWITHLIST));
    SetDlgItemTextW(hDlg, IDC_MAIN_HOTKEY_COMPLETENEXT_LABEL, GetLangStringW(wLangSystem, STRID_COMPLETENEXT));
    SetDlgItemTextW(hDlg, IDC_MAIN_HOTKEY_COMPLETEPREV_LABEL, GetLangStringW(wLangSystem, STRID_COMPLETEPREV));
    SetDlgItemTextW(hDlg, IDC_MAIN_AUTOLIST_GROUP, GetLangStringW(wLangSystem, STRID_AUTOLIST));
    SetDlgItemTextW(hDlg, IDC_MAIN_AUTOLIST_ENABLE, GetLangStringW(wLangSystem, STRID_ENABLE));
    SetDlgItemTextW(hDlg, IDC_MAIN_AUTOLISTAFTER_PRELABEL, GetLangStringW(wLangSystem, STRID_AFTERCHAR_PRE));
    SetDlgItemTextW(hDlg, IDC_MAIN_AUTOLISTAFTER_POSTLABEL, GetLangStringW(wLangSystem, STRID_AFTERCHAR_POST));
    SetDlgItemTextW(hDlg, IDC_MAIN_RIGHTDELIMITERS_ENABLE, GetLangStringW(wLangSystem, STRID_RIGHTDELIMITERS));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    if (bAutoListEnable) SendMessage(hWndAutoListEnable, BM_SETCHECK, BST_CHECKED, 0);
    SendMessage(hWndAutoListAfterSpin, UDM_SETBUDDY, (WPARAM)hWndAutoListAfter, 0);
    SendMessage(hWndAutoListAfterSpin, UDM_SETRANGE, 0, MAKELONG(99, 1));
    SetDlgItemInt(hDlg, IDC_MAIN_AUTOLISTAFTER, nAutoListAfter, FALSE);

    if (bRightDelimitersEnable) SendMessage(hWndRightDelimitersEnable, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndCompleteWithList, HKM_SETHOTKEY, dwCompleteWithList, 0);
    SendMessage(hWndCompleteNext, HKM_SETHOTKEY, dwCompleteNext, 0);
    SendMessage(hWndCompletePrev, HKM_SETHOTKEY, dwCompletePrev, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompleteWithList, pfwCompleteWithList?dwCompleteWithList:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompleteNext, pfwCompleteNext?dwCompleteNext:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCompletePrev, pfwCompletePrev?dwCompletePrev:0);

    PostMessage(hDlg, WM_COMMAND, IDC_MAIN_AUTOLIST_ENABLE, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_MAIN_AUTOLIST_ENABLE)
    {
      bAutoListEnable=SendMessage(hWndAutoListEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndAutoListAfter, bAutoListEnable);
      EnableWindow(hWndAutoListAfterSpin, bAutoListEnable);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      dwCompleteWithList=(WORD)SendMessage(hWndCompleteWithList, HKM_GETHOTKEY, 0, 0);
      dwCompleteNext=(WORD)SendMessage(hWndCompleteNext, HKM_GETHOTKEY, 0, 0);
      dwCompletePrev=(WORD)SendMessage(hWndCompletePrev, HKM_GETHOTKEY, 0, 0);

      nAutoListAfter=GetDlgItemInt(hDlg, IDC_MAIN_AUTOLISTAFTER, NULL, FALSE);

      bRightDelimitersEnable=SendMessage(hWndRightDelimitersEnable, BM_GETCHECK, 0, 0);

      SaveOptionsW(OF_SETTINGS);
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        UninitMain();
        InitMain();
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewEditParentProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static BOOL bTextChanging;

  if (uMsg == AKDN_EDIT_ONFINISH)
  {
    MANUALSET *lpElement;

    if (lpElement=StackGetManual(&hManualStack, (HWND)wParam))
      StackDeleteManual(&hManualStack, lpElement);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == ID_EDIT)
    {
      if (HIWORD(wParam) == EN_CHANGE)
      {
        EDITINFO ei;
        CHARRANGE cr;

        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
        {
          if ((HWND)lParam == ei.hWndEdit)
          {
            hWndEdit=ei.hWndEdit;
            SendMessage(hWndEdit, EM_EXGETSEL, 0, (LPARAM)&cr);

            if (hWndAutoComplete)
            {
              BLOCKINFO *lpBlockInfo=NULL;
              wchar_t wszWordPart[MAX_PATH];
              int nIndex=-1;
              BOOL bClose=TRUE;
              BOOL bEmpty=FALSE;
              BOOL bAddChar=FALSE;
              BOOL bDelChar=FALSE;

              if (cr.cpMin == cr.cpMax)
              {
                if (nWindowBlockBegin == nWindowBlockEnd)
                  bEmpty=TRUE;
                if (nWindowBlockEnd == cr.cpMin - 1)
                  bAddChar=TRUE;
                if (nWindowBlockEnd == cr.cpMin + 1)
                  bDelChar=TRUE;

                if (bAddChar || (bDelChar && !bEmpty))
                {
                  if (GetEditWordPart(&lpWindowTheme->hDelimiterStack, wszWordPart, MAX_PATH, &nWindowBlockBegin, &nWindowBlockEnd))
                  {
                    if (!bAutoListEnable || nWindowBlockEnd - nWindowBlockBegin >= nAutoListAfter)
                    {
                      if (wszWordPart[0])
                      {
                        lpBlockInfo=StackGetBlock(&lpWindowTheme->hBlockStack, wszWordPart, nWindowBlockEnd - nWindowBlockBegin, NULL);
                      }
                      if (!(bAddChar && !wszWordPart[0]) &&
                           (!wszWordPart[0] || lpBlockInfo))
                      {
                        if (lpBlockInfo) nIndex=FindStringListbox(lpBlockInfo->wpWord);
                        SetSelListbox(nIndex);
                        bClose=FALSE;
                      }
                    }
                  }
                }
                if (bClose) PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
              }
            }
            else if (bAutoListEnable)
            {
              HTHEMESTACK *lpTheme;

              if (!bReadSyntaxFiles)
              {
                bReadSyntaxFiles=TRUE;
                if (bOldWindows)
                  ReadSyntaxFilesA();
                else
                  ReadSyntaxFilesW();
              }

              if (lpTheme=StackGetCurrentScheme(&hSyntaxFilesStack))
              {
                CreateAutoCompleteWindow(lpTheme, CAW_AUTOLIST|CAW_COMPLETEWINDOW);
              }
            }

            //Remove all spots if selection outside the spot
            if (lpCurrentBlockElement)
            {
              if (!GetHotSpot(lpCurrentBlockElement, hWndEdit, &cr))
              {
                StackResetHotSpot(lpCurrentBlockElement);
                lpCurrentBlockElement=NULL;
              }
            }
          }
        }
      }
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == EN_SELCHANGE)
      {
        if (hWndAutoComplete)
        {
          if (!bTextChanging)
          {
            PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGING)
      {
        bTextChanging=TRUE;
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTINSERTEND)
      {
        if (lpCurrentBlockElement)
        {
          AENTEXTINSERT *aenti=(AENTEXTINSERT *)lParam;
          HOTSPOT *lpHotSpot=NULL;
          CHARRANGE cr;

          if (aenti->crRichRange.cpMin != aenti->crRichRange.cpMax)
          {
            cr.cpMin=aenti->crRichRange.cpMin;
            cr.cpMax=aenti->crRichRange.cpMin;

            if (lpHotSpot=GetHotSpot(lpCurrentBlockElement, aenti->hdr.hwndFrom, &cr))
            {
              lpHotSpot->nHotSpotLen+=aenti->crRichRange.cpMax - aenti->crRichRange.cpMin;

              while (lpHotSpot=lpHotSpot->next)
              {
                if (aenti->crRichRange.cpMin <= lpCurrentBlockElement->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos)
                {
                  lpHotSpot->nHotSpotPos+=aenti->crRichRange.cpMax - aenti->crRichRange.cpMin;
                }
              }
            }
            else
            {
              StackResetHotSpot(lpCurrentBlockElement);
              lpCurrentBlockElement=NULL;
            }
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTDELETEBEGIN)
      {
        if (lpCurrentBlockElement)
        {
          AENTEXTDELETE *aentd=(AENTEXTDELETE *)lParam;
          HOTSPOT *lpHotSpot=NULL;

          if (aentd->crRichRange.cpMin != aentd->crRichRange.cpMax)
          {
            if (lpHotSpot=GetHotSpot(lpCurrentBlockElement, aentd->hdr.hwndFrom, &aentd->crRichRange))
            {
              lpHotSpot->nHotSpotLen-=aentd->crRichRange.cpMax - aentd->crRichRange.cpMin;

              while (lpHotSpot=lpHotSpot->next)
              {
                if (aentd->crRichRange.cpMin <= lpCurrentBlockElement->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos)
                {
                  lpHotSpot->nHotSpotPos-=aentd->crRichRange.cpMax - aentd->crRichRange.cpMin;
                }
              }
            }
            else
            {
              StackResetHotSpot(lpCurrentBlockElement);
              lpCurrentBlockElement=NULL;
            }
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGED)
      {
        bTextChanging=FALSE;
      }
    }
  }

  //Call next procedure
  if (hWnd == hMainWnd)
    return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
  else
    return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_KILLFOCUS ||
      uMsg == WM_LBUTTONDOWN)
  {
    if (hWndAutoComplete)
    {
      PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
    }
    if (uMsg == WM_KILLFOCUS)
    {
      if (lpCurrentBlockElement)
      {
        StackResetHotSpot(lpCurrentBlockElement);
        lpCurrentBlockElement=NULL;
      }
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK GetMsgProc(int code, WPARAM wParam, LPARAM lParam)
{
  if (code >= 0)
  {
    MSG *msg=(MSG *)lParam;

    if (hWndListBox)
    {
      if (msg->message == WM_MOUSEWHEEL)
      {
        SendMessage(hWndListBox, msg->message, msg->wParam, msg->lParam);
        msg->message=WM_NULL;
      }
      else if (msg->message == WM_KEYDOWN ||
               msg->message == WM_KEYUP)
      {
        if (msg->wParam == VK_UP ||
            msg->wParam == VK_DOWN ||
            msg->wParam == VK_PRIOR ||
            msg->wParam == VK_NEXT)
        {
          if (bOldWindows)
          {
            SendMessage(hWndListBox, WM_SETREDRAW, FALSE, 0);
          }

          SendMessage(hWndListBox, msg->message, msg->wParam, msg->lParam);
          msg->message=WM_NULL;

          if (bOldWindows)
          {
            SendMessage(hWndListBox, WM_SETREDRAW, TRUE, 0);
            InvalidateRect(hWndListBox, NULL, TRUE);
          }
        }
        else
        {
          if (msg->message == WM_KEYDOWN)
          {
            if (msg->wParam == VK_ESCAPE ||
                msg->wParam == VK_RETURN)
            {
              if (msg->wParam == VK_RETURN)
              {
                BLOCKINFO *lpBlockInfo;

                if (lpBlockInfo=GetBlockListbox())
                  CompleteWordPart(lpBlockInfo, nWindowBlockBegin, nWindowBlockEnd);
              }
              PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
            }
          }
          if (msg->wParam == VK_ESCAPE ||
              msg->wParam == VK_RETURN)
          {
            msg->message=WM_NULL;
          }
        }
      }
    }
  }
  return CallNextHookEx(hHook, code, wParam, lParam);
}

BOOL CALLBACK CompleteWithListProc(void *lpParameter)
{
  HTHEMESTACK *lpTheme;
  DWORD dwCreateFlags=0;
  BOOL bCatchHotkey=TRUE;

  if (!lpParameter)
    dwCreateFlags=CAW_COMPLETEONE|CAW_COMPLETEWINDOW;
  else
    dwCreateFlags=(DWORD)lpParameter;

  if (!bReadSyntaxFiles)
  {
    bReadSyntaxFiles=TRUE;

    if (bOldWindows)
      ReadSyntaxFilesA();
    else
      ReadSyntaxFilesW();
  }
  SendMessage(hWndAutoComplete, WM_CLOSE, 0, 0);

  if (lpTheme=StackGetCurrentScheme(&hSyntaxFilesStack))
  {
    if (!CreateAutoCompleteWindow(lpTheme, dwCreateFlags))
    {
      if (!(dwCreateFlags & CAW_COMPLETEWINDOW))
        bCatchHotkey=FALSE;
    }
  }
  else if (!(dwCreateFlags & CAW_COMPLETEWINDOW))
    bCatchHotkey=FALSE;

  return bCatchHotkey;
}

BOOL CALLBACK CompleteNextProc(void *lpParameter)
{
  EDITINFO ei;
  BOOL bCatchHotkey=FALSE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    hWndEdit=ei.hWndEdit;

    if (lpCurrentBlockElement)
    {
      if (!NextHotSpot(lpCurrentBlockElement, hWndEdit, FALSE))
      {
        StackResetHotSpot(lpCurrentBlockElement);
        lpCurrentBlockElement=NULL;
      }
      else bCatchHotkey=TRUE;
    }
    else
    {
      bCatchHotkey=CompleteWithListProc((void *)(CAW_COMPLETEONE|CAW_COMPLETEEXACT));
    }
  }
  return bCatchHotkey;
}

BOOL CALLBACK CompletePrevProc(void *lpParameter)
{
  EDITINFO ei;
  BOOL bCatchHotkey=FALSE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    hWndEdit=ei.hWndEdit;

    if (lpCurrentBlockElement)
    {
      if (!NextHotSpot(lpCurrentBlockElement, hWndEdit, TRUE))
      {
        StackResetHotSpot(lpCurrentBlockElement);
        lpCurrentBlockElement=NULL;
      }
      else bCatchHotkey=TRUE;
    }
  }
  return bCatchHotkey;
}

LRESULT CALLBACK AutoCompleteWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static POINT ptResizeHitOffset={0};
  static int nResizeHit=0;

  if (uMsg == WM_CREATE)
  {
    if (bOldWindows)
    {
      hWndListBox=CreateWindowA("LISTBOX",
                              NULL,
                              WS_CHILD|WS_VSCROLL|WS_VISIBLE|
                              LBS_DISABLENOSCROLL|LBS_NOINTEGRALHEIGHT,
                              0, 0, 0, 0,
                              hWnd,
                              (HMENU)IDC_LIST,
                              hInstanceDLL,
                              NULL);

      hListboxFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWndEdit, 0));
      SendMessage(hWndListBox, WM_SETFONT, (WPARAM)hListboxFont, TRUE);

      OldListBoxProc=(WNDPROC)GetWindowLongA(hWndListBox, GWL_WNDPROC);
      SetWindowLongA(hWndListBox, GWL_WNDPROC, (LONG)NewListboxProc);
    }
    else
    {
      hWndListBox=CreateWindowW(L"LISTBOX",
                              NULL,
                              WS_CHILD|WS_VSCROLL|WS_VISIBLE|
                              LBS_DISABLENOSCROLL|LBS_NOINTEGRALHEIGHT,
                              0, 0, 0, 0,
                              hWnd,
                              (HMENU)IDC_LIST,
                              hInstanceDLL,
                              NULL);

      hListboxFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWndEdit, 0));
      SendMessage(hWndListBox, WM_SETFONT, (WPARAM)hListboxFont, TRUE);

      OldListBoxProc=(WNDPROC)GetWindowLongW(hWndListBox, GWL_WNDPROC);
      SetWindowLongW(hWndListBox, GWL_WNDPROC, (LONG)NewListboxProc);
    }
  }
  else if (uMsg == WM_SIZE)
  {
    if (hWndListBox)
    {
      SetWindowPos(hWndListBox, NULL, 0, 0, LOWORD(lParam), HIWORD(lParam), SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOMOVE);
    }
    return FALSE;
  }
  else if (uMsg == WM_MOUSEACTIVATE)
  {
    return MA_NOACTIVATE;
  }
  else if (uMsg == WM_NCHITTEST)
  {
    LRESULT lResult;

    if (bOldWindows)
      lResult=DefWindowProcA(hWnd, uMsg, wParam, lParam);
    else
      lResult=DefWindowProcW(hWnd, uMsg, wParam, lParam);

    if (lResult == HTRIGHT ||
        lResult == HTBOTTOMRIGHT ||
        lResult == HTBOTTOM)
    {
      return lResult;
    }
    return HTERROR;
  }
  else if (uMsg == WM_NCLBUTTONDOWN)
  {
    RECT rc;
    POINT pt;

    GetWindowRect(hWnd, &rc);
    GetCursorPos(&pt);

    if (wParam == HTRIGHT ||
        wParam == HTBOTTOMRIGHT ||
        wParam == HTBOTTOM)
    {
      ptResizeHitOffset.x=rc.right - pt.x;
      ptResizeHitOffset.y=rc.bottom - pt.y;
    }
    else if (wParam == HTTOPRIGHT)
    {
      ptResizeHitOffset.x=rc.right - pt.x;
      ptResizeHitOffset.y=rc.top - pt.y;
    }
    else if (wParam == HTLEFT ||
             wParam == HTTOPLEFT ||
             wParam == HTTOP)
    {
      ptResizeHitOffset.x=rc.left - pt.x;
      ptResizeHitOffset.y=rc.top - pt.y;
    }
    else if (wParam == HTBOTTOMLEFT)
    {
      ptResizeHitOffset.x=rc.left - pt.x;
      ptResizeHitOffset.y=rc.bottom - pt.y;
    }
    SetCapture(hWnd);
    nResizeHit=wParam;
    return FALSE;
  }
  else if (uMsg == WM_MOUSEMOVE)
  {
    if (nResizeHit)
    {
      RECT rc;
      POINT pt;

      GetWindowRect(hWnd, &rc);
      GetCursorPos(&pt);

      pt.x+=ptResizeHitOffset.x;
      pt.y+=ptResizeHitOffset.y;

      if (nResizeHit == HTLEFT)
      {
        rc.left=pt.x;
      }
      else if (nResizeHit == HTRIGHT)
      {
        rc.right=pt.x;
      }
      else if (nResizeHit == HTTOP)
      {
        rc.top=pt.y;
      }
      else if (nResizeHit == HTTOPLEFT)
      {
        rc.top=pt.y;
        rc.left=pt.x;
      }
      else if (nResizeHit == HTTOPRIGHT)
      {
        rc.top=pt.y;
        rc.right=pt.x;
      }
      else if (nResizeHit == HTBOTTOM)
      {
        rc.bottom=pt.y;
      }
      else if (nResizeHit == HTBOTTOMLEFT)
      {
        rc.bottom=pt.y;
        rc.left=pt.x;
      }
      else if (nResizeHit == HTBOTTOMRIGHT)
      {
        rc.bottom=pt.y;
        rc.right=pt.x;
      }
      SetWindowPos(hWnd, NULL, rc.left, rc.top, max(rc.right - rc.left, SIZE_MINX), max(rc.bottom - rc.top, SIZE_MINY), SWP_NOZORDER|SWP_NOACTIVATE);
      return FALSE;
    }
  }
  else if (uMsg == WM_LBUTTONUP ||
           uMsg == WM_CANCELMODE)
  {
    if (nResizeHit)
    {
      nResizeHit=0;
      ReleaseCapture();
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    RECT rc;

    GetWindowRect(hWnd, &rc);
    rc.right-=rc.left;
    rc.bottom-=rc.top;
    if (rcAutoComplete.right != rc.right ||
        rcAutoComplete.bottom != rc.bottom)
    {
      rcAutoComplete=rc;
      bAutoCompleteWindowRectSave=TRUE;
    }
    CloseAutoCompleteWindow();
  }
  else if (uMsg == WM_DESTROY)
  {
    hWndAutoComplete=NULL;
    hWndListBox=NULL;
  }

  if (bOldWindows)
    return DefWindowProcA(hWnd, uMsg, wParam, lParam);
  else
    return DefWindowProcW(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewListboxProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_MOUSEACTIVATE)
  {
    return MA_NOACTIVATE;
  }
  else if (uMsg == WM_LBUTTONDOWN)
  {
    LRESULT lResult=SendMessage(hWnd, LB_ITEMFROMPOINT, 0, lParam);

    if (HIWORD(lResult) == 0)
    {
      SendMessage(hWnd, LB_SETCURSEL, LOWORD(lResult), 0);
    }
    return FALSE;
  }
  else if (uMsg == WM_LBUTTONUP)
  {
    return FALSE;
  }
  else if (uMsg == WM_MBUTTONDOWN)
  {
    return FALSE;
  }
  else if (uMsg == WM_MBUTTONUP)
  {
    return FALSE;
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    BLOCKINFO *lpBlockInfo;

    if (lpBlockInfo=GetBlockListbox())
      CompleteWordPart(lpBlockInfo, nWindowBlockBegin, nWindowBlockEnd);
    PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
    return FALSE;
  }

  if (bOldWindows)
    return CallWindowProcA(OldListBoxProc, hWnd, uMsg, wParam, lParam);
  else
    return CallWindowProcW(OldListBoxProc, hWnd, uMsg, wParam, lParam);
}

BOOL CreateAutoCompleteWindow(HTHEMESTACK *lpTheme, DWORD dwFlags)
{
  BLOCKINFO *lpBlockInfo=NULL;
  wchar_t wszWordPart[MAX_PATH];
  int nIndex=-1;
  int nBlockBegin;
  int nBlockEnd;
  BOOL bOnlyOne=FALSE;

  if (!GetEditWordPart(&lpTheme->hDelimiterStack, wszWordPart, MAX_PATH, &nBlockBegin, &nBlockEnd))
    return FALSE;

  if (dwFlags & CAW_AUTOLIST)
  {
    if (nBlockEnd - nBlockBegin < nAutoListAfter)
      return FALSE;
  }

  //Complete without window
  if (wszWordPart[0])
  {
    if (lpBlockInfo=StackGetBlock(&lpTheme->hBlockStack, wszWordPart, nBlockEnd - nBlockBegin, &bOnlyOne))
    {
      if (bOnlyOne || ((dwFlags & CAW_COMPLETEEXACT) && !xstrcmpiW(lpBlockInfo->wpWord, wszWordPart)))
      {
        if (dwFlags & CAW_COMPLETEONE)
        {
          CompleteWordPart(lpBlockInfo, nBlockBegin, nBlockEnd);
          return TRUE;
        }
      }
    }
  }

  //Complete with window
  if (dwFlags & CAW_COMPLETEWINDOW)
  {
    if (!wszWordPart[0] || lpBlockInfo)
    {
      if (bOldWindows)
      {
        hWndAutoComplete=CreateWindowA(AUTOCOMPLETEA,
                              NULL,
                              WS_POPUP|WS_THICKFRAME,
                              0, 0, rcAutoComplete.right, rcAutoComplete.bottom,
                              hMainWnd,
                              NULL,
                              hInstanceDLL,
                              NULL);
      }
      else
      {
        hWndAutoComplete=CreateWindowW(AUTOCOMPLETEW,
                              NULL,
                              WS_POPUP|WS_THICKFRAME,
                              0, 0, rcAutoComplete.right, rcAutoComplete.bottom,
                              hMainWnd,
                              NULL,
                              hInstanceDLL,
                              NULL);
      }
      lpWindowTheme=lpTheme;
      nWindowBlockBegin=nBlockBegin;
      nWindowBlockEnd=nBlockEnd;

      if (MoveAutoCompleteWindow())
      {
        FillListbox(&lpTheme->hBlockStack);
        if (lpBlockInfo) nIndex=FindStringListbox(lpBlockInfo->wpWord);
        SetSelListbox(nIndex);
        ShowWindow(hWndAutoComplete, SW_SHOWNOACTIVATE);

        hHook=SetWindowsHookEx(WH_GETMESSAGE, GetMsgProc, NULL, GetCurrentThreadId());
      }
      else PostMessage(hWndAutoComplete, WM_CLOSE, 0, 0);
    }
  }
  return FALSE;
}

BOOL MoveAutoCompleteWindow()
{
  RECT rc;
  POINT pt;
  int nCharHeight;

  if (!hWndAutoComplete) return FALSE;
  GetPosFromChar(hWndEdit, nWindowBlockBegin, &pt);
  nCharHeight=GetCharHeight(hWndEdit);
  pt.y+=nCharHeight;

  if (ClientToScreen(hWndEdit, &pt))
  {
    GetWindowRect(hWndEdit, &rc);

    if (pt.x + rcAutoComplete.right > rc.right &&
        pt.x - rcAutoComplete.right >= rc.left)
    {
      pt.x=pt.x - rcAutoComplete.right;
    }
    if (pt.y + rcAutoComplete.bottom > rc.bottom &&
        pt.y - rcAutoComplete.bottom - nCharHeight >= rc.top)
    {
      pt.y=pt.y - rcAutoComplete.bottom - nCharHeight;
    }
    if (pt.x >= 0 && pt.y >= 0 && pt.x < rc.right && pt.y < rc.bottom)
    {
      SetWindowPos(hWndAutoComplete, NULL, pt.x, pt.y, 0, 0, SWP_NOZORDER|SWP_NOSIZE|SWP_NOACTIVATE);
      return TRUE;
    }
  }
  return FALSE;
}

void CloseAutoCompleteWindow()
{
  if (hListboxFont)
  {
    DeleteObject(hListboxFont);
    hListboxFont=NULL;
  }
  if (hHook)
  {
    if (UnhookWindowsHookEx(hHook))
      hHook=NULL;
  }
}

void FillListbox(HBLOCKSTACK *hBlockStack)
{
  BLOCKINFO *lpBlockElement=(BLOCKINFO *)hBlockStack->first;
  char szWord[MAX_PATH];
  int nItem;

  while (lpBlockElement)
  {
    if (bOldWindows)
    {
      WideCharToMultiByte(CP_ACP, 0, lpBlockElement->wpWord, lpBlockElement->nWordLen + 1, szWord, MAX_PATH, NULL, NULL);
      if ((nItem=SendMessageA(hWndListBox, LB_INSERTSTRING, (WPARAM)-1, (LPARAM)szWord)) != LB_ERR)
        SendMessage(hWndListBox, LB_SETITEMDATA, nItem, (LPARAM)lpBlockElement);
    }
    else
    {
      if ((nItem=SendMessageW(hWndListBox, LB_INSERTSTRING, (WPARAM)-1, (LPARAM)lpBlockElement->wpWord)) != LB_ERR)
        SendMessage(hWndListBox, LB_SETITEMDATA, nItem, (LPARAM)lpBlockElement);
    }

    lpBlockElement=lpBlockElement->next;
  }
}

int FindStringListbox(wchar_t *wpWord)
{
  char szWord[MAX_PATH];

  if (bOldWindows)
  {
    WideCharToMultiByte(CP_ACP, 0, wpWord, -1, szWord, MAX_PATH, NULL, NULL);
    return SendMessageA(hWndListBox, LB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)szWord);
  }
  else return SendMessageW(hWndListBox, LB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)wpWord);
}

void SetSelListbox(int nIndex)
{
  if (bOldWindows)
  {
    SendMessage(hWndListBox, WM_SETREDRAW, FALSE, 0);
  }
  if (nIndex != -1)
  {
    SendMessage(hWndListBox, LB_SETCURSEL, nIndex, 0);
    SendMessage(hWndListBox, LB_SETTOPINDEX, nIndex, 0);
  }
  else
  {
    SendMessage(hWndListBox, LB_SETCURSEL, 0, 0);
    SendMessage(hWndListBox, LB_SETCURSEL, (WPARAM)-1, 0);
    SendMessage(hWndListBox, LB_SETTOPINDEX, 0, 0);
  }
  if (bOldWindows)
  {
    SendMessage(hWndListBox, WM_SETREDRAW, TRUE, 0);
    InvalidateRect(hWndListBox, NULL, TRUE);
  }
}

BLOCKINFO* GetBlockListbox()
{
  int nItem;

  if ((nItem=SendMessage(hWndListBox, LB_GETCURSEL, 0, 0)) != LB_ERR)
  {
    return (BLOCKINFO *)SendMessage(hWndListBox, LB_GETITEMDATA, nItem, 0);
  }
  return NULL;
}

HTHEMESTACK* StackLoadSyntaxFile(HANDLE hFile, HSTACK *hSyntaxFilesStack)
{
  HTHEMESTACK *lpScheme=NULL;
  EXTENSIONINFO *lpExtElement=NULL;
  DELIMITERINFO *lpDelimElement=NULL;
  VARIABLEINFO *lpVarElement=NULL;
  wchar_t *wszText;
  wchar_t *wpText;
  wchar_t *wpExtension;
  wchar_t wchDelimiter;
  wchar_t *wpVarName;
  wchar_t *wpVarValue;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  int nExtensionLen;
  int nVarNameLen;
  int nVarValueLen;
  int nWordLen=0;
  int nBlockLen;
  int nBlockParsedLen;
  int nLinesInBlock;
  BOOL bQuoteString;

  if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
  {
    if (dwFileSize >= 2)
    {
      if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
      {
        if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
        {
          dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
          wszText[dwUnicodeLen++]='\0';
          if (wszText[0] == 0xFEFF)
          {
            ++wpText;
            --dwUnicodeLen;
          }

          ////SCHEME
          if (!(lpScheme=StackInsertScheme(hSyntaxFilesStack)))
            return NULL;

          while (1)
          {
            SectionStart:
            SkipCommentW(&wpText);
            GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL);
            if (!NextLineW(&wpText)) goto FreeText;

            if (!xstrcmpiW(wszBuffer, L"Extensions:"))
            {
              while (1)
              {
                //Parse line
                while (1)
                {
                  lpExtElement=NULL;
                  wpExtension=NULL;

                  //Extension
                  if ((nExtensionLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                  {
                    if (!bQuoteString && wszBuffer[0] == ';') break;

                    if (wpExtension=(wchar_t *)GlobalAlloc(GPTR, nExtensionLen * sizeof(wchar_t) + 2))
                      memcpy(wpExtension, wszBuffer, nExtensionLen * sizeof(wchar_t) + 2);
                    else
                      break;
                  }
                  else goto SectionStart;

                  //Add to stack
                  if (lpExtElement=StackInsertExtension(&lpScheme->hExtensionStack, nExtensionLen))
                  {
                    lpExtElement->wpExtension=wpExtension;
                    lpExtElement->nExtensionLen=nExtensionLen;
                  }
                  break;
                }
                if (!lpExtElement && wpExtension) GlobalFree((HGLOBAL)wpExtension);
                if (!NextLineW(&wpText)) goto FreeText;
              }
            }
            else if (!xstrcmpiW(wszBuffer, L"Delimiters:"))
            {
              while (1)
              {
                //Parse line
                if (nWordLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                {
                  wchDelimiter=*wszBuffer;
                }
                else goto SectionStart;

                //Add to stack
                if (lpDelimElement=StackInsertDelimiter(&lpScheme->hDelimiterStack))
                  lpDelimElement->wchDelimiter=wchDelimiter;

                //Next line
                if (!NextLineW(&wpText)) goto FreeText;
              }
            }
            else if (!xstrcmpiW(wszBuffer, L"Variables:"))
            {
              while (1)
              {
                //Parse line
                while (1)
                {
                  lpVarElement=NULL;
                  wpVarName=NULL;
                  wpVarValue=NULL;

                  //Var name
                  if (nVarNameLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                  {
                    if (wszBuffer[0] == ';') break;

                    if (wpVarName=(wchar_t *)GlobalAlloc(GPTR, nVarNameLen * sizeof(wchar_t) + 2))
                      memcpy(wpVarName, wszBuffer, nVarNameLen * sizeof(wchar_t) + 2);
                    else
                      break;
                  }
                  else goto SectionStart;

                  //Var value
                  if (nVarValueLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                  {
                    if (wpVarValue=(wchar_t *)GlobalAlloc(GPTR, nVarValueLen * sizeof(wchar_t) + 2))
                      memcpy(wpVarValue, wszBuffer, nVarValueLen * sizeof(wchar_t) + 2);
                    else
                      break;
                  }
                  else break;

                  //Add to stack
                  if (lpVarElement=StackInsertVariable(&lpScheme->hVarStack))
                  {
                    lpVarElement->wpVarName=wpVarName;
                    lpVarElement->nVarNameLen=nVarNameLen;
                    lpVarElement->wpVarValue=wpVarValue;
                    lpVarElement->nVarValueLen=nVarValueLen;
                  }
                  break;
                }
                if (!lpVarElement)
                {
                  if (wpVarName) GlobalFree((HGLOBAL)wpVarName);
                  if (wpVarValue) GlobalFree((HGLOBAL)wpVarValue);
                }
                if (!NextLineW(&wpText)) goto FreeText;
              }
            }
            else if (!xstrcmpiW(wszBuffer, L"Words:"))
            {
              BLOCKINFO *lpBlockMaster=NULL;
              BLOCKINFO *lpBlockInfo=NULL;
              WORDINFO *lpWordInfo=NULL;
              wchar_t *wpWordBegin;
              wchar_t *wszWord;
              wchar_t *wpBlockBegin;
              wchar_t *wpBlockEnd;
              wchar_t *wszBlockParsed;
              HSTACK hWordStack;
              HSTACK hHotSpotStack;

              while (*wpText)
              {
                lpBlockMaster=NULL;
                lpBlockInfo=NULL;
                lpWordInfo=NULL;
                wpWordBegin=NULL;
                wszWord=NULL;
                wszBlockParsed=NULL;
                hWordStack.first=0;
                hWordStack.last=0;
                hHotSpotStack.first=0;
                hHotSpotStack.last=0;

                //Word
                while (*wpText)
                {
                  //Skip comment
                  if (*wpText == '$' && (*(wpText + 1) == ';' || *(wpText + 1) == '#') && *(wpText - 1) != '$')
                    while (*++wpText != '\r' && *wpText != '\0');

                  if (*wpText == '$' && *(wpText + 1) == '~' && *(wpText - 1) != '$')
                  {
                    wpText+=2;
                    wpWordBegin=wpText;
                    while (*wpText != ' ' && *wpText != '\t' && *wpText != '\r' && *wpText != '\0') ++wpText;
                    nWordLen=wpText - wpWordBegin;

                    if (nWordLen)
                    {
                      if (wszWord=(wchar_t *)GlobalAlloc(GPTR, nWordLen * sizeof(wchar_t) + 2))
                      {
                        memcpy(wszWord, wpWordBegin, nWordLen * sizeof(wchar_t));
                        wszWord[nWordLen]='\0';

                        if (lpWordInfo=StackInsertWord(&hWordStack))
                        {
                          lpWordInfo->wpWord=wszWord;
                          lpWordInfo->nWordLen=nWordLen;
                        }
                        else break;

                        wszWord=NULL;
                      }
                      else break;
                    }
                  }
                  if (*wpText == '\r')
                  {
                    NextLineW(&wpText);
                    break;
                  }
                  ++wpText;
                }

                //Word block
                if (*wpText && hWordStack.first)
                {
                  wpBlockBegin=wpText;
                  while (*wpText)
                  {
                    //Skip comment
                    if (*wpText == '$' && (*(wpText + 1) == ';' || *(wpText + 1) == '#') && *(wpText - 1) != '$')
                      while (*++wpText != '\r' && *wpText != '\0');

                    //Find next abbreviation definition
                    if (*wpText == '$' && *(wpText + 1) == '~' && *(wpText - 1) != '$')
                      break;
                    ++wpText;
                  }
                  wpBlockEnd=wpText - 1;
                  nBlockLen=(wpBlockEnd - wpBlockBegin) + 1;

                  if (nBlockParsedLen=ParseBlock(lpScheme, &hHotSpotStack, wpBlockBegin, nBlockLen, NULL, NULL))
                  {
                    if (wszBlockParsed=(wchar_t *)GlobalAlloc(GPTR, nBlockParsedLen * sizeof(wchar_t)))
                    {
                      nBlockParsedLen=ParseBlock(lpScheme, &hHotSpotStack, wpBlockBegin, nBlockLen, wszBlockParsed, &nLinesInBlock);
                      if (nBlockParsedLen > 0 && wszBlockParsed[nBlockParsedLen - 1] == '\r') wszBlockParsed[--nBlockParsedLen]='\0';
                      if (nBlockParsedLen > 0 && wszBlockParsed[nBlockParsedLen - 1] == '\r') wszBlockParsed[--nBlockParsedLen]='\0';

                      for (lpWordInfo=(WORDINFO *)hWordStack.first; lpWordInfo; lpWordInfo=lpWordInfo->next)
                      {
                        if (lpBlockInfo=StackInsertBlock(&lpScheme->hBlockStack, lpWordInfo->wpWord, lpWordInfo->nWordLen))
                        {
                          lpBlockInfo->wpWord=lpWordInfo->wpWord;
                          lpBlockInfo->nWordLen=lpWordInfo->nWordLen;

                          if (lpWordInfo == (WORDINFO *)hWordStack.first)
                          {
                            lpBlockMaster=lpBlockInfo;
                            lpBlockInfo->master=NULL;
                            lpBlockInfo->wpBlock=wszBlockParsed;
                            lpBlockInfo->nBlockLen=nBlockParsedLen;
                            lpBlockInfo->nLinesInBlock=nLinesInBlock;
                            lpBlockInfo->hHotSpotStack=hHotSpotStack;
                          }
                          else lpBlockInfo->master=lpBlockMaster;
                        }
                      }
                      wszBlockParsed=NULL;
                    }
                  }
                }

                //Free
                if (!lpBlockInfo)
                  StackFreeHotSpot(&hHotSpotStack);
                StackFreeWord(&hWordStack);
                if (wszWord) GlobalFree((HGLOBAL)wszWord);
                if (wszBlockParsed) GlobalFree((HGLOBAL)wszBlockParsed);
              }
            }
            else break;
          }
        }
        FreeText:
        GlobalFree((HGLOBAL)wszText);
      }
    }
  }
  return lpScheme;
}

int ParseBlock(HTHEMESTACK *lpScheme, HSTACK *hHotSpotStack, const wchar_t *wpInput, int nInputLen, wchar_t *wszOutput, int *nOutputLines)
{
  VARIABLEINFO *lpElement;
  HOTSPOT *lpHotSpot=NULL;
  const wchar_t *wpVarNameStart;
  int nLines=1;
  int nVarNameLen;
  int a;
  int b;

  for (a=0, b=0; b < nInputLen; ++b)
  {
    if (wpInput[b] == '$')
    {
      if (wpInput[++b] == ';' || wpInput[b] == '#')
      {
        while (++b < nInputLen && wpInput[b] != '\r' && wpInput[b] != '\0');
      }
      else if (wpInput[b] == '$')
      {
        if (wszOutput) wszOutput[a]='$';
        ++a;
        if (lpHotSpot) ++lpHotSpot->nHotSpotInitLen;
      }
      else if (wpInput[b] == '\\')
      {
        if (++b >= nInputLen) break;
        if (wszOutput) wszOutput[a]=wpInput[b];
        ++a;
        if (lpHotSpot) ++lpHotSpot->nHotSpotInitLen;
      }
      else if (wpInput[b] == '[')
      {
        if (wszOutput) lpHotSpot=StackInsertHotSpot(hHotSpotStack, a);
      }
      else if (wpInput[b] == '{')
      {
        wpVarNameStart=wpInput + b + 1;
        for (nVarNameLen=0; ++b < nInputLen && wpInput[b] != '}'; ++nVarNameLen);
        if (b >= nInputLen) break;

        if (lpElement=StackGetVariable(&lpScheme->hVarStack, wpVarNameStart, nVarNameLen))
        {
          if (wszOutput) memcpy(wszOutput + a, lpElement->wpVarValue, lpElement->nVarValueLen * sizeof(wchar_t));
          a+=lpElement->nVarValueLen;
          if (lpHotSpot) lpHotSpot->nHotSpotInitLen+=lpElement->nVarValueLen;
        }
      }
    }
    else if (wpInput[b] == ']')
    {
      if (lpHotSpot)
      {
        lpHotSpot->nHotSpotLen=lpHotSpot->nHotSpotInitLen;
        lpHotSpot=NULL;
      }
    }
    else if (wpInput[b] == '\n')
    {
      //Skip '\n'
    }
    else
    {
      if (wpInput[b] == '\r') ++nLines;
      if (wszOutput) wszOutput[a]=wpInput[b];
      ++a;
      if (lpHotSpot) ++lpHotSpot->nHotSpotInitLen;
    }
  }
  if (wszOutput)
    wszOutput[a]='\0';
  else
    ++a;
  if (nOutputLines) *nOutputLines=nLines;
  return a;
}

HTHEMESTACK* StackInsertScheme(HSTACK *hStack)
{
  HTHEMESTACK *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HTHEMESTACK));
  return lpElement;
}

HTHEMESTACK* StackGetScheme(HSTACK *hStack, const wchar_t *wpExtension, int nExtensionLen)
{
  HTHEMESTACK *lpElement=(HTHEMESTACK *)hStack->first;

  while (lpElement)
  {
    if (StackGetExtension(&lpElement->hExtensionStack, wpExtension, nExtensionLen))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

HTHEMESTACK* StackGetCurrentScheme(HSTACK *hStack)
{
  HTHEMESTACK *hTheme=NULL;
  MANUALSET *lpManual;
  EDITINFO ei;
  wchar_t wszExt[MAX_PATH];
  wchar_t *wpExt=NULL;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    hWndEdit=ei.hWndEdit;

    if (lpManual=StackGetManual(&hManualStack, hWndEdit))
    {
      wpExt=lpManual->wszExt;
    }
    else
    {
      if (bOldWindows)
      {
        MultiByteToWideChar(CP_ACP, 0, GetFileExtA((char *)ei.pFile), -1, wszExt, MAX_PATH);
        wpExt=wszExt;
      }
      else wpExt=GetFileExtW((wchar_t *)ei.pFile);
    }

    if (wpExt)
    {
      hTheme=StackGetScheme(hStack, wpExt, lstrlenW(wpExt));
    }
  }
  return hTheme;
}

void StackFreeScheme(HSTACK *hStack)
{
  HTHEMESTACK *lpElement=(HTHEMESTACK *)hStack->first;

  while (lpElement)
  {
    StackFreeExtension(&lpElement->hExtensionStack);
    StackFreeDelimiter(&lpElement->hDelimiterStack);
    StackFreeVariable(&lpElement->hVarStack);
    StackFreeBlock(&lpElement->hBlockStack);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

EXTENSIONINFO* StackInsertExtension(HEXTENSIONSTACK *hStack, int nExtensionLen)
{
  EXTENSIONINFO *lpElement;
  EXTENSIONINFO *lpElementNew=NULL;

  if (hStack->lpExtensionLens[nExtensionLen])
  {
    lpElement=(EXTENSIONINFO *)hStack->lpExtensionLens[nExtensionLen];
  }
  else
  {
    lpElement=(EXTENSIONINFO *)hStack->first;

    while (lpElement)
    {
      if (lpElement->nExtensionLen >= nExtensionLen)
        break;

      lpElement=lpElement->next;
    }
  }
  StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement, (stack **)&lpElementNew, sizeof(EXTENSIONINFO));

  if (lpElementNew)
    hStack->lpExtensionLens[nExtensionLen]=(int)lpElementNew;
  return lpElementNew;
}

EXTENSIONINFO* StackGetExtension(HEXTENSIONSTACK *hStack, const wchar_t *wpExtension, int nExtensionLen)
{
  EXTENSIONINFO *lpElement=(EXTENSIONINFO *)hStack->lpExtensionLens[nExtensionLen];

  while (lpElement)
  {
    if (lpElement->nExtensionLen != nExtensionLen)
      break;
    if (!xstrcmpiW(lpElement->wpExtension, wpExtension))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeExtension(HEXTENSIONSTACK *hStack)
{
  EXTENSIONINFO *lpElement=(EXTENSIONINFO *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpExtension);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

DELIMITERINFO* StackInsertDelimiter(HSTACK *hStack)
{
  DELIMITERINFO *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(DELIMITERINFO));
  return lpElement;
}

DELIMITERINFO* StackGetDelimiter(HSTACK *hStack, wchar_t wchDelimiter)
{
  DELIMITERINFO *lpElement=(DELIMITERINFO *)hStack->first;

  while (lpElement)
  {
    if (lpElement->wchDelimiter == wchDelimiter)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeDelimiter(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

VARIABLEINFO* StackInsertVariable(HSTACK *hStack)
{
  VARIABLEINFO *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(VARIABLEINFO));
  return lpElement;
}

VARIABLEINFO* StackGetVariable(HSTACK *hStack, const wchar_t *wpVarName, int nVarNameLen)
{
  VARIABLEINFO *lpElement=(VARIABLEINFO *)hStack->first;

  while (lpElement)
  {
    if (lpElement->nVarNameLen == nVarNameLen)
    {
      if (!xstrcmpinW(wpVarName, lpElement->wpVarName, nVarNameLen))
        return lpElement;
    }
    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeVariable(HSTACK *hStack)
{
  VARIABLEINFO *lpElement=(VARIABLEINFO *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpVarName);
    GlobalFree((HGLOBAL)lpElement->wpVarValue);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

WORDINFO* StackInsertWord(HSTACK *hStack)
{
  WORDINFO *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(WORDINFO));
  return lpElement;
}

void StackFreeWord(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

BOOL GetEditWordPart(HSTACK *hDelimiterStack, wchar_t *wszWord, int nWordMax, int *nMin, int *nMax)
{
  GETTEXTRANGE gtr;
  CHARRANGE cr;
  wchar_t *wpWord;
  wchar_t wchRightChar;
  int nLine;
  int nLineIndex;
  int nRangeLength;
  int nCount=0;
  int nLen;
  BOOL bBreak=FALSE;
  BOOL bResult=FALSE;

  SendMessage(hWndEdit, EM_EXGETSEL, 0, (LPARAM)&cr);

  if (cr.cpMin == cr.cpMax)
  {
    //Check right delimiter
    if (bRightDelimitersEnable)
    {
      gtr.cpMin=cr.cpMin;
      gtr.cpMax=cr.cpMin + 1;

      if (SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWndEdit, (LPARAM)&gtr))
      {
        if (bOldWindows)
          MultiByteToWideChar(CP_ACP, 0, (char *)gtr.pText, 1, &wchRightChar, 1);
        else
          wchRightChar=*(wchar_t *)gtr.pText;

        if (wchRightChar != ' ' &&
            wchRightChar != '\t' &&
            wchRightChar != '\r' &&
            !StackGetDelimiter(hDelimiterStack, wchRightChar))
        {
          bBreak=TRUE;
        }
        SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
      }
    }
    if (bBreak) return FALSE;

    //Get left word part
    nLine=SendMessage(hWndEdit, EM_EXLINEFROMCHAR, 0, cr.cpMin);
    nLineIndex=SendMessage(hWndEdit, EM_LINEINDEX, nLine, 0);
    nRangeLength=cr.cpMin - nLineIndex;

    if (nRangeLength > 0 && nRangeLength < MAX_PATH)
    {
      gtr.cpMin=nLineIndex;
      gtr.cpMax=nLineIndex + nRangeLength;

      if (nLen=SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWndEdit, (LPARAM)&gtr))
      {
        if (bOldWindows)
        {
          wpWord=(wchar_t *)GlobalAlloc(GPTR, nLen * sizeof(wchar_t) + 2);
          MultiByteToWideChar(CP_ACP, 0, (char *)gtr.pText, nLen + 1, wpWord, nLen + 1);
        }
        else wpWord=(wchar_t *)gtr.pText;

        for (nCount=nRangeLength - 1; nCount >= 0; --nCount)
        {
          if (wpWord[nCount] == ' ' ||
              wpWord[nCount] == '\t' ||
              wpWord[nCount] == '\r' ||
              StackGetDelimiter(hDelimiterStack, wpWord[nCount]))
          {
            break;
          }
        }
        if (++nCount != nRangeLength)
        {
          xstrcpynW(wszWord, wpWord + nCount, nWordMax);
          *nMin=nLineIndex + nCount;
          *nMax=nLineIndex + nRangeLength;
          bResult=TRUE;
        }

        if (bOldWindows)
          GlobalFree((HGLOBAL)wpWord);
        SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
      }
    }
    if (nCount == nRangeLength)
    {
      wszWord[0]='\0';
      *nMin=cr.cpMin;
      *nMax=cr.cpMin;
      bResult=TRUE;
    }
  }
  return bResult;
}

void CompleteWordPart(BLOCKINFO *lpBlockInfo, int nMin, int nMax)
{
  HOTSPOT *lpHotSpot;
  GETTEXTRANGE gtr;
  CHARRANGE cr;
  EDITINFO ei;
  wchar_t *wpLine;
  wchar_t *wpIndentBlock=NULL;
  int nMinLine;
  int nMinLineIndex;
  int nLen;
  int nSpaceCount=0;
  int a;
  int b;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (!ei.bReadOnly)
    {
      if (bAkelEdit)
      {
        if (lpCurrentBlockElement)
          StackResetHotSpot(lpCurrentBlockElement);
        lpCurrentBlockElement=NULL;
      }
      if (lpBlockInfo->master) lpBlockInfo=lpBlockInfo->master;

      if (lpBlockInfo->nLinesInBlock > 1)
      {
        nMinLine=SendMessage(hWndEdit, EM_EXLINEFROMCHAR, 0, nMin);
        nMinLineIndex=SendMessage(hWndEdit, EM_LINEINDEX, nMinLine, 0);

        if (nMinLineIndex != nMin)
        {
          gtr.cpMin=nMinLineIndex;
          gtr.cpMax=nMin;

          //Get line before word
          if (nLen=SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWndEdit, (LPARAM)&gtr))
          {
            if (bOldWindows)
            {
              wpLine=(wchar_t *)GlobalAlloc(GPTR, nLen * sizeof(wchar_t) + 2);
              MultiByteToWideChar(CP_ACP, 0, (char *)gtr.pText, nLen + 1, wpLine, nLen + 1);
            }
            else wpLine=(wchar_t *)gtr.pText;

            //Get spaces
            for (nSpaceCount=0; wpLine[nSpaceCount] == ' ' || wpLine[nSpaceCount] == '\t'; ++nSpaceCount);
            wpLine[nSpaceCount]='\0';

            if (nSpaceCount)
            {
              nLen=lpBlockInfo->nBlockLen + (lpBlockInfo->nLinesInBlock - 1) * nSpaceCount;

              if (wpIndentBlock=(wchar_t *)GlobalAlloc(GPTR, nLen * sizeof(wchar_t) + 2))
              {
                for (a=0, b=0; b < lpBlockInfo->nBlockLen;)
                {
                  if (lpBlockInfo->wpBlock[b] == '\r')
                  {
                    wpIndentBlock[a++]=lpBlockInfo->wpBlock[b++];

                    for (lpHotSpot=(HOTSPOT *)lpBlockInfo->hHotSpotStack.first; lpHotSpot; lpHotSpot=lpHotSpot->next)
                    {
                      if (lpHotSpot->nHotSpotPos >= a)
                      {
                        lpHotSpot->nHotSpotPos+=nSpaceCount;
                      }
                    }
                    xstrcpyW(wpIndentBlock + a, wpLine);
                    a+=nSpaceCount;
                  }
                  else wpIndentBlock[a++]=lpBlockInfo->wpBlock[b++];
                }
                wpIndentBlock[a]='\0';
              }
            }

            if (bOldWindows)
              GlobalFree((HGLOBAL)wpLine);
            SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
          }
        }
      }
      if (!wpIndentBlock)
        wpIndentBlock=lpBlockInfo->wpBlock;

      cr.cpMin=nMin;
      cr.cpMax=nMax;
      SendMessage(hWndEdit, EM_EXSETSEL, 0, (LPARAM)&cr);
      SendMessage(hMainWnd, AKD_REPLACESELW, (WPARAM)hWndEdit, (LPARAM)wpIndentBlock);

      if (lpHotSpot=(HOTSPOT *)lpBlockInfo->hHotSpotStack.first)
      {
        cr.cpMin=nMin + lpHotSpot->nHotSpotPos;
        cr.cpMax=nMin + lpHotSpot->nHotSpotPos + lpHotSpot->nHotSpotLen;
        SendMessage(hWndEdit, EM_EXSETSEL, 0, (LPARAM)&cr);
      }
      if (bAkelEdit)
      {
        lpBlockInfo->nHotSpotBlockBegin=nMin;
        lpCurrentBlockElement=lpBlockInfo;
      }

      if (wpIndentBlock != lpBlockInfo->wpBlock)
        GlobalFree((HGLOBAL)wpIndentBlock);
    }
    else MessageBeep(MB_OK);
  }
}

BLOCKINFO* StackInsertBlock(HBLOCKSTACK *hBlockStack, wchar_t *wpWord, int nWordLen)
{
  BLOCKINFO *lpElement=(BLOCKINFO *)hBlockStack->first;
  BLOCKINFO *lpElementNew=NULL;
  wchar_t wchFirstUpperChar=WideCharUpper(*wpWord);

  if (wchFirstUpperChar < FIRST_NONLATIN)
  {
    if (hBlockStack->lpSorted[wchFirstUpperChar])
      lpElement=(BLOCKINFO *)hBlockStack->lpSorted[wchFirstUpperChar];
    else
      lpElement=(BLOCKINFO *)hBlockStack->first;
  }
  else lpElement=(BLOCKINFO *)hBlockStack->lpSorted[FIRST_NONLATIN];

  while (lpElement)
  {
    if (lpElement->wchFirstUpperChar >= wchFirstUpperChar)
    {
      if (xstrcmpiW(lpElement->wpWord, wpWord) >= 0)
        break;
    }
    lpElement=lpElement->next;
  }
  StackInsertBefore((stack **)&hBlockStack->first, (stack **)&hBlockStack->last, (stack *)lpElement, (stack **)&lpElementNew, sizeof(BLOCKINFO));

  if (lpElementNew)
  {
    if (wchFirstUpperChar < FIRST_NONLATIN)
    {
      if (!hBlockStack->lpSorted[wchFirstUpperChar] || (int)lpElement == hBlockStack->lpSorted[wchFirstUpperChar])
        hBlockStack->lpSorted[wchFirstUpperChar]=(int)lpElementNew;
    }
    else
    {
      if (hBlockStack->lpSorted[FIRST_NONLATIN] || (int)lpElement == hBlockStack->lpSorted[FIRST_NONLATIN])
        hBlockStack->lpSorted[FIRST_NONLATIN]=(int)lpElementNew;
    }
    lpElementNew->wchFirstUpperChar=wchFirstUpperChar;
    lpElementNew->wpWord=wpWord;
    lpElementNew->nWordLen=nWordLen;
  }
  return lpElementNew;
}

BLOCKINFO* StackGetBlock(HBLOCKSTACK *hBlockStack, const wchar_t *wpWordPart, int nWordPartLen, BOOL *bOnlyOne)
{
  BLOCKINFO *lpBlockElement;
  wchar_t wchFirstUpperChar=WideCharUpper(*wpWordPart);

  if (wchFirstUpperChar < FIRST_NONLATIN)
    lpBlockElement=(BLOCKINFO *)hBlockStack->lpSorted[wchFirstUpperChar];
  else
    lpBlockElement=(BLOCKINFO *)hBlockStack->lpSorted[FIRST_NONLATIN];

  while (lpBlockElement)
  {
    if (wchFirstUpperChar < FIRST_NONLATIN && wchFirstUpperChar != lpBlockElement->wchFirstUpperChar)
      break;

    if (nWordPartLen <= lpBlockElement->nWordLen)
    {
      if (!xstrcmpinW(wpWordPart, lpBlockElement->wpWord, nWordPartLen))
      {
        if (bOnlyOne)
        {
          if (lpBlockElement->next && !xstrcmpinW(wpWordPart, lpBlockElement->next->wpWord, nWordPartLen))
            *bOnlyOne=FALSE;
          else
            *bOnlyOne=TRUE;
        }
        return lpBlockElement;
      }
    }
    lpBlockElement=lpBlockElement->next;
  }
  return NULL;
}

void StackFreeBlock(HBLOCKSTACK *hBlockStack)
{
  BLOCKINFO *lpElement=(BLOCKINFO *)hBlockStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpWord);
    if (!lpElement->master)
    {
      GlobalFree((HGLOBAL)lpElement->wpBlock);
      StackFreeHotSpot(&lpElement->hHotSpotStack);
    }

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hBlockStack->first, (stack **)&hBlockStack->last);
}

HOTSPOT* StackInsertHotSpot(HSTACK *hStack, int nHotSpotPos)
{
  HOTSPOT *lpElement=NULL;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HOTSPOT)))
  {
    lpElement->nHotSpotInitPos=nHotSpotPos;
    lpElement->nHotSpotInitLen=0;
    lpElement->nHotSpotPos=nHotSpotPos;
    lpElement->nHotSpotLen=0;
  }
  return lpElement;
}

HOTSPOT* GetHotSpot(BLOCKINFO *lpBlockInfo, HWND hWnd, CHARRANGE *cr)
{
  HOTSPOT *lpHotSpot=(HOTSPOT *)lpBlockInfo->hHotSpotStack.first;

  while (lpHotSpot)
  {
    if (lpHotSpot->nHotSpotPos >= 0)
    {
      if (cr->cpMin >= lpBlockInfo->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos &&
          cr->cpMax <= lpBlockInfo->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos + lpHotSpot->nHotSpotLen)
      {
        break;
      }
    }
    lpHotSpot=lpHotSpot->next;
  }
  return lpHotSpot;
}

BOOL NextHotSpot(BLOCKINFO *lpBlockInfo, HWND hWnd, BOOL bPrev)
{
  HOTSPOT *lpHotSpot;
  CHARRANGE crOld;
  CHARRANGE crNew;

  SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&crOld);

  if (lpHotSpot=GetHotSpot(lpBlockInfo, hWnd, &crOld))
  {
    if (!bPrev)
    {
      if (lpHotSpot && lpHotSpot->next)
        lpHotSpot=lpHotSpot->next;
      else
        lpHotSpot=(HOTSPOT *)lpBlockInfo->hHotSpotStack.first;
    }
    else
    {
      if (lpHotSpot && lpHotSpot->prev)
        lpHotSpot=lpHotSpot->prev;
      else
        lpHotSpot=(HOTSPOT *)lpBlockInfo->hHotSpotStack.last;
    }

    if (lpHotSpot)
    {
      crNew.cpMin=lpBlockInfo->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos;
      crNew.cpMax=lpBlockInfo->nHotSpotBlockBegin + lpHotSpot->nHotSpotPos + lpHotSpot->nHotSpotLen;
      if (crOld.cpMin != crNew.cpMin || crOld.cpMax != crNew.cpMax)
      {
        SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&crNew);
        return TRUE;
      }
    }
  }
  return FALSE;
}

void StackResetHotSpot(BLOCKINFO *lpBlockInfo)
{
  HOTSPOT *lpElement=(HOTSPOT *)lpBlockInfo->hHotSpotStack.first;

  while (lpElement)
  {
    lpElement->nHotSpotPos=lpElement->nHotSpotInitPos;
    lpElement->nHotSpotLen=lpElement->nHotSpotInitLen;

    lpElement=lpElement->next;
  }
}

void StackFreeHotSpot(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

MANUALSET* StackInsertManual(HSTACK *hStack)
{
  MANUALSET *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(MANUALSET));

  return lpElement;
}

MANUALSET* StackGetManual(HSTACK *hStack, HWND hWnd)
{
  MANUALSET *lpElement=(MANUALSET *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackDeleteManual(HSTACK *hStack, void *ms)
{
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)ms);
}

int GetCharHeight(HWND hWnd)
{
  HDC hDC;
  HFONT hFont;
  HFONT hFontOld;
  int nHeight=0;

  if (hDC=GetDC(hWnd))
  {
    if (bOldWindows)
    {
      TEXTMETRICA tm;

      hFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      hFontOld=(HFONT)SelectObject(hDC, hFont);
      GetTextMetricsA(hDC, &tm);
      nHeight=tm.tmHeight;
      if (hFontOld) SelectObject(hDC, hFontOld);
      DeleteObject(hFont);
    }
    else
    {
      TEXTMETRICW tm;

      hFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      hFontOld=(HFONT)SelectObject(hDC, hFont);
      GetTextMetricsW(hDC, &tm);
      nHeight=tm.tmHeight;
      if (hFontOld) SelectObject(hDC, hFontOld);
      DeleteObject(hFont);
    }
    ReleaseDC(hWnd, hDC);
  }
  return nHeight;
}

void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt)
{
  DWORD dwPosition;

  if (bOldRichEdit)
  {
    dwPosition=SendMessage(hWnd, EM_POSFROMCHAR, nCharIndex, (LPARAM)NULL);
    pt->x=LOWORD(dwPosition);
    pt->y=HIWORD(dwPosition);
  }
  else
  {
    SendMessage(hWnd, EM_POSFROMCHAR, (WPARAM)pt, nCharIndex);
  }
}

int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote)
{
  char *pCount;
  char chStopChar;
  int nWordLen;

  while (*pText == ' ' || *pText == '\t') ++pText;

  if (*pText == '\"' || *pText == '\'' || *pText == '`')
  {
    //Parse: "param" or 'param' or `param`
    chStopChar=*pText;
    pCount=++pText;
    while (*pCount != chStopChar && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
      if (*pCount == chStopChar)
        ++*pNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    pCount=pText;

    while (*pCount != ' ' && *pCount != '\t' && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(pCount - pText, nWordLenMax - 1);
  memcpy(szWord, pText, nWordLen);
  szWord[nWordLen]='\0';

  return nWordLen;
}

int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote)
{
  wchar_t *wpCount;
  wchar_t wchStopChar;
  int nWordLen;

  while (*wpText == ' ' || *wpText == '\t') ++wpText;

  if (*wpText == '\"' || *wpText == '\'' || *wpText == '`')
  {
    //Parse: "param" or 'param' or `param`
    wchStopChar=*wpText;
    wpCount=++wpText;
    while (*wpCount != wchStopChar && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
      if (*wpCount == wchStopChar)
        ++*wpNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    wpCount=wpText;

    while (*wpCount != ' ' && *wpCount != '\t' && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(wpCount - wpText, nWordLenMax - 1);
  memcpy(wszWord, wpText, nWordLen * sizeof(wchar_t));
  wszWord[nWordLen]='\0';

  return nWordLen;
}

BOOL NextLineA(char **pText)
{
  while (**pText != '\r' && **pText != '\0') ++*pText;
  if (**pText == '\0') return FALSE;
  if (*++*pText == '\n') ++*pText;
  return TRUE;
}

BOOL NextLineW(wchar_t **wpText)
{
  while (**wpText != '\r' && **wpText != '\0') ++*wpText;
  if (**wpText == '\0') return FALSE;
  if (*++*wpText == '\n') ++*wpText;
  return TRUE;
}

BOOL SkipCommentA(char **pText)
{
  while (1)
  {
    while (**pText == ' ' || **pText == '\t' || **pText == '\r' || **pText == '\n') ++*pText;

    if (**pText == ';')
    {
      while (**pText != '\r' && **pText != '\0') ++*pText;
    }
    else break;
  }
  if (**pText == '\0')
    return FALSE;
  return TRUE;
}

BOOL SkipCommentW(wchar_t **wpText)
{
  while (1)
  {
    while (**wpText == ' ' || **wpText == '\t' || **wpText == '\r' || **wpText == '\n') ++*wpText;

    if (**wpText == ';')
    {
      while (**wpText != '\r' && **wpText != '\0') ++*wpText;
    }
    else break;
  }
  if (**wpText == '\0')
    return FALSE;
  return TRUE;
}

char* GetFileExtA(char *pFile)
{
  int i;

  for (i=lstrlenA(pFile) - 1; i >= 0; --i)
  {
    if (pFile[i] == '.') return (pFile + i + 1);
    else if (pFile[i] == '\\') break;
  }
  return "";
}

wchar_t* GetFileExtW(wchar_t *wpFile)
{
  int i;

  for (i=lstrlenW(wpFile) - 1; i >= 0; --i)
  {
    if (wpFile[i] == '.') return (wpFile + i + 1);
    else if (wpFile[i] == '\\') break;
  }
  return L"";
}


//// Options

void ReadSyntaxFilesA()
{
  WIN32_FIND_DATAA wfd;
  HANDLE hSearch;
  HANDLE hFile;
  char szFindFiles[MAX_PATH];
  char szAutoCompleteFile[MAX_PATH];

  wsprintfA(szFindFiles, "%s\\*.autocomplete", szAutoCompleteDir);

  if ((hSearch=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      wsprintfA(szAutoCompleteFile, "%s\\%s", szAutoCompleteDir, wfd.cFileName);
      if ((hFile=CreateFileA(szAutoCompleteFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
      {
        StackLoadSyntaxFile(hFile, &hSyntaxFilesStack);
        CloseHandle(hFile);
      }
    }
    while (FindNextFileA(hSearch, &wfd));

    FindClose(hSearch);
  }
}

void ReadSyntaxFilesW()
{
  WIN32_FIND_DATAW wfd;
  HANDLE hSearch;
  HANDLE hFile;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszAutoCompleteFile[MAX_PATH];

  wsprintfW(wszFindFiles, L"%s\\*.autocomplete", wszAutoCompleteDir);

  if ((hSearch=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      if (wfd.cFileName[0] == '.' && (wfd.cFileName[1] == '\0' || (wfd.cFileName[1] == '.' && wfd.cFileName[2] == '\0'))) continue;

      wsprintfW(wszAutoCompleteFile, L"%s\\%s", wszAutoCompleteDir, wfd.cFileName);
      if ((hFile=CreateFileW(wszAutoCompleteFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
      {
        StackLoadSyntaxFile(hFile, &hSyntaxFilesStack);
        CloseHandle(hFile);
      }
    }
    while (FindNextFileW(hSearch, &wfd));

    FindClose(hSearch);
  }
}

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="CompleteWithList";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompleteWithList;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CompleteNext";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompleteNext;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CompletePrev";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompletePrev;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoListEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bAutoListEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AutoListAfter";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nAutoListAfter;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="RightDelimitersEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bRightDelimitersEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcAutoComplete;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"CompleteWithList";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompleteWithList;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CompleteNext";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompleteNext;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CompletePrev";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCompletePrev;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoListEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bAutoListEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AutoListAfter";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nAutoListAfter;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"RightDelimitersEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bRightDelimitersEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcAutoComplete;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName="CompleteWithList";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompleteWithList;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CompleteNext";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompleteNext;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CompletePrev";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompletePrev;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoListEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bAutoListEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AutoListAfter";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nAutoListAfter;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="RightDelimitersEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bRightDelimitersEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcAutoComplete;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName=L"CompleteWithList";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompleteWithList;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CompleteNext";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompleteNext;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CompletePrev";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCompletePrev;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoListEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bAutoListEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AutoListAfter";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nAutoListAfter;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"RightDelimitersEnable";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bRightDelimitersEnable;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcAutoComplete;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_HOTKEYS)
      return L"\x0413\x043E\x0440\x044F\x0447\x0438\x0435\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0438";
    if (nStringID == STRID_COMPLETEWITHLIST)
      return L"\x0414\x043E\x043F\x043E\x043B\x043D\x0438\x0442\x044C\x0020\x0441\x043E\x0020\x0441\x043F\x0438\x0441\x043A\x043E\x043C";
    if (nStringID == STRID_COMPLETENEXT)
      return L"\x0414\x043E\x043F\x043E\x043B\x043D\x0438\x0442\x044C\x0020\x0431\x0435\x0437\x0020\x0441\x043F\x0438\x0441\x043A\x0430\x0020\x002F\n\x041A\x0020\x0441\x043B\x0435\x0434\x0443\x044E\x0449\x0435\x0439\x0020\x043F\x043E\x0437\x0438\x0446\x0438\x0438";
    if (nStringID == STRID_COMPLETEPREV)
      return L"\x041A\x0020\x043F\x0440\x0435\x0434\x044B\x0434\x0443\x044E\x0449\x0435\x0439\x0020\x043F\x043E\x0437\x0438\x0446\x0438\x0438";
    if (nStringID == STRID_AUTOLIST)
      return L"\x0410\x0432\x0442\x043E\x043C\x0430\x0442\x0438\x0447\x0435\x0441\x043A\x0438\x0439\x0020\x0441\x043F\x0438\x0441\x043E\x043A";
    if (nStringID == STRID_ENABLE)
      return L"\x0412\x043A\x043B\x044E\x0447\x0435\x043D\x043E";
    if (nStringID == STRID_AFTERCHAR_PRE)
      return L"\x041F\x043E\x043A\x0430\x0437\x0430\x0442\x044C\x0020\x043F\x043E\x0441\x043B\x0435";
    if (nStringID == STRID_AFTERCHAR_POST)
      return L"\x0441\x0438\x043C\x0432\x043E\x043B\x043E\x0432";
    if (nStringID == STRID_RIGHTDELIMITERS)
      return L"\x0423\x0447\x0438\x0442\x044B\x0432\x0430\x0442\x044C\x0020\x0440\x0430\x0437\x0434\x0435\x043B\x0438\x0442\x0435\x043B\x0438\x0020\x0441\x043F\x0440\x0430\x0432\x0430\x0020\x043E\x0442\x0020\x043A\x0430\x0440\x0435\x0442\x043A\x0438";
    if (nStringID == STRID_OK)
      return L"\x041E\x041A";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_HOTKEYS)
      return L"Hotkeys";
    if (nStringID == STRID_COMPLETEWITHLIST)
      return L"Complete with list";
    if (nStringID == STRID_COMPLETENEXT)
      return L"Complete without list /\nGo to next position";
    if (nStringID == STRID_COMPLETEPREV)
      return L"Go to previous position";
    if (nStringID == STRID_AUTOLIST)
      return L"Automatic list";
    if (nStringID == STRID_ENABLE)
      return L"Enable";
    if (nStringID == STRID_AFTERCHAR_PRE)
      return L"Show after";
    if (nStringID == STRID_AFTERCHAR_POST)
      return L"characters";
    if (nStringID == STRID_RIGHTDELIMITERS)
      return L"Consider delimiters from right at the caret";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  bOldRichEdit=pd->bOldRichEdit;
  bAkelEdit=pd->bAkelEdit;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szAutoCompleteDir, "%s\\AkelFiles\\Plugs\\AutoComplete", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszAutoCompleteDir, L"%s\\AkelFiles\\Plugs\\AutoComplete", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  //Add hotkeys
  if (bOldWindows)
  {
    PLUGINFUNCTIONA pf={0};

    if (dwCompleteWithList && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompleteWithList))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "AutoComplete::Main::WithList");
      pf.wHotkey=(WORD)dwCompleteWithList;
      pf.PluginProc=CompleteWithListProc;
      pfaCompleteWithList=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCompleteNext && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompleteNext))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "AutoComplete::Main::Next");
      pf.wHotkey=(WORD)dwCompleteNext;
      pf.PluginProc=CompleteNextProc;
      pfaCompleteNext=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCompletePrev && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompletePrev))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "AutoComplete::Main::Prev");
      pf.wHotkey=(WORD)dwCompletePrev;
      pf.PluginProc=CompletePrevProc;
      pfaCompletePrev=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }
  else
  {
    PLUGINFUNCTIONW pf={0};

    if (dwCompleteWithList && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompleteWithList))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"AutoComplete::Main::WithList");
      pf.wHotkey=(WORD)dwCompleteWithList;
      pf.PluginProc=CompleteWithListProc;
      pfwCompleteWithList=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCompleteNext && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompleteNext))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"AutoComplete::Main::Next");
      pf.wHotkey=(WORD)dwCompleteNext;
      pf.PluginProc=CompleteNextProc;
      pfwCompleteNext=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCompletePrev && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCompletePrev))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"AutoComplete::Main::Prev");
      pf.wHotkey=(WORD)dwCompletePrev;
      pf.PluginProc=CompletePrevProc;
      pfwCompletePrev=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }

  //Subclass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewEditParentProc, (LPARAM)&NewFrameProcData);
  }

  NewEditProcData=NULL;
  SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProc, (LPARAM)&NewEditProcData);

  if (bOldWindows)
  {
    WNDCLASSA wndclass={0};

    wndclass.style        =CS_HREDRAW|CS_VREDRAW;
    wndclass.lpfnWndProc  =AutoCompleteWindowProc;
    wndclass.cbClsExtra   =0;
    wndclass.cbWndExtra   =DLGWINDOWEXTRA;
    wndclass.hInstance    =hInstanceDLL;
    wndclass.hIcon        =NULL;
    wndclass.hCursor      =LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground=(HBRUSH)GetStockObject(HOLLOW_BRUSH);
    wndclass.lpszMenuName =NULL;
    wndclass.lpszClassName=AUTOCOMPLETEA;
    RegisterClassA(&wndclass);
  }
  else
  {
    WNDCLASSW wndclass={0};

    wndclass.style        =CS_HREDRAW|CS_VREDRAW;
    wndclass.lpfnWndProc  =AutoCompleteWindowProc;
    wndclass.cbClsExtra   =0;
    wndclass.cbWndExtra   =DLGWINDOWEXTRA;
    wndclass.hInstance    =hInstanceDLL;
    wndclass.hIcon        =NULL;
    wndclass.hCursor      =LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground=(HBRUSH)GetStockObject(HOLLOW_BRUSH);
    wndclass.lpszMenuName =NULL;
    wndclass.lpszClassName=AUTOCOMPLETEW;
    RegisterClassW(&wndclass);
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove hotkeys
  if (bOldWindows)
  {
    if (pfaCompleteWithList)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCompleteWithList);
      pfaCompleteWithList=NULL;
    }
    if (pfaCompleteNext)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCompleteNext);
      pfaCompleteNext=NULL;
    }
    if (pfaCompletePrev)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCompletePrev);
      pfaCompletePrev=NULL;
    }
  }
  else
  {
    if (pfwCompleteWithList)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCompleteWithList);
      pfwCompleteWithList=NULL;
    }
    if (pfwCompleteNext)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCompleteNext);
      pfwCompleteNext=NULL;
    }
    if (pfwCompletePrev)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCompletePrev);
      pfwCompletePrev=NULL;
    }
  }

  //Remove subclass
  if (IsWindow(hMainWnd))
  {
    if (NewMainProcData)
    {
      SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
      NewMainProcData=NULL;
    }
    if (bMDI)
    {
      if (NewFrameProcData)
      {
        SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
        NewFrameProcData=NULL;
      }
    }
    if (NewEditProcData)
    {
      SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
      NewEditProcData=NULL;
    }
  }

  CloseAutoCompleteWindow();
  StackFreeScheme(&hSyntaxFilesStack);
  bReadSyntaxFiles=FALSE;

  if (bOldWindows)
  {
    UnregisterClassA(AUTOCOMPLETEA, hInstanceDLL);
    if (bAutoCompleteWindowRectSave) SaveOptionsA(OF_WINDOWRECT);
  }
  else
  {
    UnregisterClassW(AUTOCOMPLETEW, hInstanceDLL);
    if (bAutoCompleteWindowRectSave) SaveOptionsW(OF_WINDOWRECT);
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
